@@ -5,22 +5,36 @@ from aiohttp import web
 from cbpi.api.step import CBPiStep, StepResult
 from cbpi.api.timer import Timer
 from cbpi.api.dataclasses import Kettle, Props
-import datetime
+from datetime import datetime
 import time
 from cbpi.api import *
 import logging
+from socket import timeout
+from typing import KeysView
+from cbpi.api.config import ConfigType
+from cbpi.api.base import CBPiBase
+from voluptuous.schema_builder import message
+from cbpi.api.dataclasses import NotificationAction, NotificationType
+import numpy
+import requests
 
 @parameters([Property.Text(label="Notification",configurable = True, description = "Text for notification"),
              Property.Select(label="AutoNext",options=["Yes","No"], description="Automatically move to next step (Yes) or pause after Notification (No)")])
 
 class BM_SimpleStep(CBPiStep):
 
+    async def NextStep(self, **kwargs):
+        await self.next()
+
     async def on_timer_done(self,timer):
         self.summary = self.props.Notification
-        self.cbpi.notify(self.props.Notification)
-        await self.push_update()
+
         if self.AutoNext == True:
+            self.cbpi.notify(self.name, self.props.Notification, NotificationType.INFO)
             await self.next()
+        else:
+            self.cbpi.notify(self.name, self.props.Notification, NotificationType.INFO, action=[NotificationAction("Next Step", self.NextStep)])
+            await self.push_update()
 
     async def on_timer_update(self,timer, seconds):
         await self.push_update()
@@ -30,7 +44,6 @@ class BM_SimpleStep(CBPiStep):
         self.AutoNext = False if self.props.AutoNext == "No" else True
         if self.timer is None:
             self.timer = Timer(1 ,on_update=self.on_timer_update, on_done=self.on_timer_done)
-        self.timer.start()
         await self.push_update()
 
     async def on_stop(self):
@@ -39,8 +52,12 @@ class BM_SimpleStep(CBPiStep):
         await self.push_update()
 
     async def run(self):
-        while True:
-           await asyncio.sleep(1)
+        while self.running == True:
+            await asyncio.sleep(1)
+            if self.timer.is_running is not True:
+                self.timer.start()
+                self.timer.is_running = True
+
         return StepResult.DONE
 
 @parameters([Property.Number(label="Temp", configurable=True),
@@ -50,17 +67,21 @@ class BM_SimpleStep(CBPiStep):
 
 class BM_MashInStep(CBPiStep):
 
+    async def NextStep(self, **kwargs):
+        await self.next()
+
     async def on_timer_done(self,timer):
         self.summary = "MashIn Temp reached. Please add Malt Pipe."
-        self.cbpi.notify("MashIn Temp reached. Please add malt pipe and malt. Move to next step")
         await self.push_update()
         if self.AutoMode == True:
             await self.setAutoMode(False)
+        self.cbpi.notify(self.name, 'MashIn Temp reached. Please add malt pipe and malt. Move to next step', action=[NotificationAction("Next Step", self.NextStep)])
 
     async def on_timer_update(self,timer, seconds):
         await self.push_update()
 
     async def on_start(self):
+        self.port = str(self.cbpi.static_config.get('port',8000))
         self.AutoMode = True if self.props.AutoMode == "Yes" else False
         self.kettle=self.get_kettle(self.props.Kettle)
         self.kettle.target_temp = int(self.props.Temp)
@@ -79,19 +100,19 @@ class BM_MashInStep(CBPiStep):
         await self.push_update()
 
     async def run(self):
-        while True:
+        while self.running == True:
            await asyncio.sleep(1)
            sensor_value = self.get_sensor_value(self.props.Sensor).get("value")
-           if sensor_value >= int(self.props.Temp) and self.timer._task == None:
+           if sensor_value >= int(self.props.Temp) and self.timer.is_running is not True:
                self.timer.start()
+               self.timer.is_running = True
         await self.push_update()
         return StepResult.DONE
 
     async def setAutoMode(self, auto_state):
         try:
-            #todo: get port from config
             if (self.kettle.instance is None or self.kettle.instance.state == False) and (auto_state is True):
-                url="http://127.0.0.1:8000/kettle/"+ self.kettle.id+"/toggle"
+                url="http://127.0.0.1:" + self.port + "/kettle/"+ self.kettle.id+"/toggle"
                 async with aiohttp.ClientSession() as session:
                     async with session.post(url) as response:
                         return await response.text()
@@ -115,24 +136,26 @@ class BM_MashStep(CBPiStep):
     @action("Start Timer", [])
     async def start_timer(self):
         if self.timer._task == None:
-            self.cbpi.notify("Timer started")
+            self.cbpi.notify(self.name, 'Timer started', NotificationType.INFO)
             self.timer.start()
         else:
-            self.cbpi.notify("Timer is already running")
+            self.cbpi.notify(self.name, 'Timer is already running', NotificationType.WARNING)
 
     @action("Add 5 Minutes to Timer", [])
     async def add_timer(self):
         if self.timer._task != None:
-            self.cbpi.notify("5 Minutes added")
+            self.cbpi.notify(self.name, '5 Minutes added', NotificationType.INFO)
             await self.timer.add(300)       
         else:
-            self.cbpi.notify("Timer must be running to add time")
+            self.cbpi.notify(self.name, 'Timer must be running to add time', NotificationType.WARNING)
 
 
     async def on_timer_done(self,timer):
         self.summary = ""
         if self.AutoMode == True:
             await self.setAutoMode(False)
+        self.cbpi.notify(self.name, 'Step finished', NotificationType.SUCCESS)
+       
         await self.next()
 
     async def on_timer_update(self,timer, seconds):
@@ -140,6 +163,7 @@ class BM_MashStep(CBPiStep):
         await self.push_update()
 
     async def on_start(self):
+        self.port = str(self.cbpi.static_config.get('port',8000))
         self.AutoMode = True if self.props.AutoMode == "Yes" else False
         self.kettle=self.get_kettle(self.props.Kettle)
         self.kettle.target_temp = int(self.props.Temp)
@@ -163,19 +187,20 @@ class BM_MashStep(CBPiStep):
         self.timer = Timer(int(self.props.Timer) *60 ,on_update=self.on_timer_update, on_done=self.on_timer_done)
 
     async def run(self):
-        while True:
+        while self.running == True:
             await asyncio.sleep(1)
             sensor_value = self.get_sensor_value(self.props.Sensor).get("value")
-            if sensor_value >= int(self.props.Temp) and self.timer._task == None:
+            if sensor_value >= int(self.props.Temp) and self.timer.is_running is not True:
                 self.timer.start()
-                self.cbpi.notify("Timer started")
+                self.timer.is_running = True
+                estimated_completion_time = datetime.fromtimestamp(time.time()+ (int(self.props.Timer))*60)
+                self.cbpi.notify(self.name, 'Timer started. Estimated completion: {}'.format(estimated_completion_time.strftime("%H:%M")), NotificationType.INFO)
         return StepResult.DONE
 
     async def setAutoMode(self, auto_state):
         try:
-            #todo: get port from config
             if (self.kettle.instance is None or self.kettle.instance.state == False) and (auto_state is True):
-                url="http://127.0.0.1:8000/kettle/"+ self.kettle.id+"/toggle"
+                url="http://127.0.0.1:" + self.port + "/kettle/"+ self.kettle.id+"/toggle"
                 async with aiohttp.ClientSession() as session:
                     async with session.post(url) as response:
                         return await response.text()
@@ -189,40 +214,86 @@ class BM_MashStep(CBPiStep):
             logging.error("Failed to switch on KettleLogic {} {}".format(self.kettle.id, e))
 
 
-@parameters([Property.Number(label="Temp", configurable=True),
-             Property.Sensor(label="Sensor"),
+@parameters([Property.Number(label="Temp", configurable=True, description="Target temperature for cooldown. Notification will be send when temp is reached and Actor can be triggered"),
+             Property.Number(label="Interval", configurable=True, description="Interval [min] for Notification and caclulate predicted End time"),
+             Property.Sensor(label="Sensor", description="Sensor that is used during cooldown"),
+             Property.Actor(label="Actor", description="Actor can trigger a valve for the cooldwon to target temperature"),
              Property.Kettle(label="Kettle")])
 class BM_Cooldown(CBPiStep):
 
     async def on_timer_done(self,timer):
         self.summary = ""
-        self.cbpi.notify("Wort cooled down. Please transfer to Fermenter.")
+        if self.actor is not None:
+            await self.actor_off(self.actor)
+        self.cbpi.notify('CoolDown', 'Wort cooled down. Please transfer to Fermenter.', NotificationType.INFO)
         await self.next()
 
     async def on_timer_update(self,timer, seconds):
         await self.push_update()
 
     async def on_start(self):
+        self.temp_array = []
+        self.time_array = []
         self.kettle = self.get_kettle(self.props.Kettle)
-        self.target_temp= int(self.props.Temp)
-        self.summary="Cool down to {}°".format(self.target_temp)
-        self.cbpi.notify("Cool down to {}°".format(self.target_temp))
+        self.actor= None
+        if self.props.Actor or self.props.Actor is not None or self.props.Actor !="":
+            self.actor = self.props.Actor
+
+        self.target_temp = int(self.props.Temp)
+        try:
+            self.Interval = int(self.props.Interval) # Interval on how often cooldwon end time is calculated
+        except:
+            self.Interval = 10
+
+        self.cbpi.notify(self.name, 'Cool down to {}°'.format(self.target_temp), NotificationType.INFO)
         if self.timer is None:
             self.timer = Timer(1,on_update=self.on_timer_update, on_done=self.on_timer_done)
+        self.start_time=time.time()
+        self.temp_array.append(self.get_sensor_value(self.props.Sensor).get("value"))
+        self.time_array.append(time.time())
+        self.next_check = self.start_time + self.Interval * 60
+        self.count = 0
 
     async def on_stop(self):
         await self.timer.stop()
         self.summary = ""
+        if self.actor is not None:
+            await self.actor_off(self.actor)
         await self.push_update()
 
     async def reset(self):
         self.timer = Timer(1,on_update=self.on_timer_update, on_done=self.on_timer_done)
 
     async def run(self):
-        while True:
-            if self.get_sensor_value(self.props.Sensor).get("value") <= self.target_temp:
+        timestring = datetime.fromtimestamp(self.start_time)
+        if self.actor is not None:
+            await self.actor_on(self.actor)
+        self.summary="Cool down to {}° started: {}".format(self.target_temp, timestring.strftime("%H:%M"))
+        await self.push_update()
+        while self.running == True:
+            current_temp = self.get_sensor_value(self.props.Sensor).get("value")
+            if self.count == 29:
+                self.temp_array.append(current_temp)
+                self.time_array.append(time.time())
+                self.count = 0
+                logging.info(self.temp_array)
+                logging.info(self.time_array)
+            if time.time() >= self.next_check:
+                self.next_check = time.time() + (self.Interval * 60)
+                cooldown_model = numpy.poly1d(numpy.polyfit(self.temp_array, self.time_array, 2))
+                target_time=cooldown_model(self.target_temp)
+                target_timestring= datetime.fromtimestamp(target_time)
+                self.summary="Cool down to {}° ECD: {}".format(self.target_temp, target_timestring.strftime("%d.%m %H:%M:%S"))
+                self.cbpi.notify("Cooldown Step","Current: {}°, reaching {}° at {}".format(round(current_temp,1), self.target_temp, target_timestring.strftime("%d.%m %H:%M")), NotificationType.INFO)
+                await self.push_update()
+
+            if current_temp <= self.target_temp and self.timer.is_running is not True:
                 self.timer.start()
+                self.timer.is_running = True
+
+            self.count +=1
             await asyncio.sleep(1)
+
         return StepResult.DONE
 
 @parameters([Property.Number(label="Timer", description="Time in Minutes", configurable=True), 
@@ -235,31 +306,33 @@ class BM_Cooldown(CBPiStep):
              Property.Number("Hop_2", configurable=True, description="Second Hop alert (minutes before finish)"),
              Property.Number("Hop_3", configurable=True, description="Third Hop alert (minutes before finish)"),
              Property.Number("Hop_4", configurable=True, description="Fourth Hop alert (minutes before finish)"),
-             Property.Number("Hop_5", configurable=True, description="Fifth Hop alert (minutes before finish)")])
+             Property.Number("Hop_5", configurable=True, description="Fifth Hop alert (minutes before finish)"),
+             Property.Number("Hop_6", configurable=True, description="Sixth Hop alert (minutes before finish)")])
 
 class BM_BoilStep(CBPiStep):
 
     @action("Start Timer", [])
     async def start_timer(self):
         if self.timer._task == None:
-            self.cbpi.notify("Timer started")
+            self.cbpi.notify(self.name, 'Timer started', NotificationType.INFO)
             self.timer.start()
         else:
-            self.cbpi.notify("Timer is already running")
+            self.cbpi.notify(self.name, 'Timer is already running', NotificationType.WARNING)
 
     @action("Add 5 Minutes to Timer", [])
     async def add_timer(self):
         if self.timer._task != None:
-            self.cbpi.notify("5 Minutes added")
+            self.cbpi.notify(self.name, '5 Minutes added', NotificationType.INFO)
             await self.timer.add(300)       
         else:
-            self.cbpi.notify("Timer must be running to add time")
+            self.cbpi.notify(self.name, 'Timer must be running to add time', NotificationType.WARNING)
 
     async def on_timer_done(self,timer):
         self.summary = ""
         self.kettle.target_temp = 0
         if self.AutoMode == True:
             await self.setAutoMode(False)
+        self.cbpi.notify(self.name, 'Boiling completed', NotificationType.SUCCESS)
         await self.next()
 
     async def on_timer_update(self,timer, seconds):
@@ -268,10 +341,13 @@ class BM_BoilStep(CBPiStep):
         await self.push_update()
 
     async def on_start(self):
+        self.lid_temp = 95 if self.get_config_value("TEMP_UNIT", "C") == "C" else 203
+        self.lid_flag = False
+        self.port = str(self.cbpi.static_config.get('port',8000))
         self.AutoMode = True if self.props.AutoMode == "Yes" else False
         self.first_wort_hop_flag = False 
         self.first_wort_hop=self.props.First_Wort 
-        self.hops_added=["","","","",""]
+        self.hops_added=["","","","","",""]
         self.remaining_seconds = None
 
         self.kettle=self.get_kettle(self.props.Kettle)
@@ -286,10 +362,10 @@ class BM_BoilStep(CBPiStep):
         await self.push_update()
 
     async def check_hop_timer(self, number, value):
-        if value is not None and self.hops_added[number-1] is not True:
+        if value is not None and value != '' and self.hops_added[number-1] is not True:
             if self.remaining_seconds != None and self.remaining_seconds <= (int(value) * 60 + 1):
                 self.hops_added[number-1]= True
-                self.cbpi.notify("Hop Alert. Please add Hop %s" % number)
+                self.cbpi.notify('Hop Alert', "Please add Hop %s" % number, NotificationType.INFO)
 
     async def on_stop(self):
         await self.timer.stop()
@@ -305,26 +381,35 @@ class BM_BoilStep(CBPiStep):
     async def run(self):
         if self.first_wort_hop_flag == False and self.first_wort_hop == "Yes":
             self.first_wort_hop_flag = True
-            self.cbpi.notify("First Wort Hop Addition! Please add hops for first wort")
+            self.cbpi.notify('First Wort Hop Addition!', 'Please add hops for first wort', NotificationType.INFO)
 
-        while True:
+        while self.running == True:
             await asyncio.sleep(1)
-            sensor_value = self.get_sensor_value(self.props.Sensor)
-            if sensor_value is not None and sensor_value.get("value") >= int(self.props.Temp) and self.timer._task == None:
+            sensor_value = self.get_sensor_value(self.props.Sensor).get("value")
+            
+            if self.lid_flag == False and sensor_value >= self.lid_temp:
+                self.cbpi.notify("Please remove lid!", "Reached temp close to boiling", NotificationType.INFO)
+                self.lid_flag = True
+
+            if sensor_value >= int(self.props.Temp) and self.timer.is_running is not True:
                 self.timer.start()
+                self.timer.is_running = True
+                estimated_completion_time = datetime.fromtimestamp(time.time()+ (int(self.props.Timer))*60)
+                self.cbpi.notify(self.name, 'Timer started. Estimated completion: {}'.format(estimated_completion_time.strftime("%H:%M")), NotificationType.INFO)
             else:
                 await self.check_hop_timer(1, self.props.Hop_1)
                 await self.check_hop_timer(2, self.props.Hop_2)
                 await self.check_hop_timer(3, self.props.Hop_3)
                 await self.check_hop_timer(4, self.props.Hop_4)
                 await self.check_hop_timer(5, self.props.Hop_5)
+                await self.check_hop_timer(6, self.props.Hop_6)
+
         return StepResult.DONE
 
     async def setAutoMode(self, auto_state):
         try:
-            #todo: get port from config
             if (self.kettle.instance is None or self.kettle.instance.state == False) and (auto_state is True):
-                url="http://127.0.0.1:8000/kettle/"+ self.kettle.id+"/toggle"
+                url="http://127.0.0.1:" + self.port + "/kettle/" + self.kettle.id + "/toggle"
                 async with aiohttp.ClientSession() as session:
                     async with session.post(url) as response:
                        return await response.text()
